﻿XIncludeFile("PB64__DECLARE.pbi")

Macro VIC_IsSpriteEnabled(n_)
	ISSET_BIT(VIC\Sprite_Enabled, n_)
EndMacro

Macro VIC_IsMulticolorSprite(n_)
	ISSET_BIT(VIC\Sprite_Multicolor, n_)
EndMacro

Macro VIC_IsDoubleHeightSprite(n_)
	ISSET_BIT(VIC\Sprite_DoubleHeight, n_)
EndMacro

Macro VIC_IsDoubleWidthSprite(n_)
	ISSET_BIT(VIC\Sprite_DoubleWidth, n_)
EndMacro

Macro VIC_SpritePriority(n_)
	ISSET_BIT(vic\Sprite_Priority, n_)
EndMacro	

Procedure VIC_Init()
	Protected i
	
	ClearStructure(@VIC, S_VIC)
	InitializeStructure(@VIC, S_VIC)
	
	VIC\NextRasterLine			= #LineCycles
	VIC\Display_Enabled			= 1
	
	; Default memory pointers
	VIC\Mem_Screen				= $0400
	VIC\Mem_Char				= $D000
	VIC\Mem_Bitmap				= $0000
	VIC\Mem_VIC_Base			= 1
	VIC\GraphicMode 			= #STANDARD_TEXT
	VIC\Cycle					= 1
EndProcedure

Procedure VIC_RasterIRQ()	
	VIC\IRQ_Flag | $1
	If VIC\IRQ_Mask & $01
		VIC\IRQ_Flag | $80
		CPU\IRQ_State | #VIC_IRQ
	EndIf
EndProcedure

Procedure.a VIC_ReadRegister(r.a)
	Protected retval.a
	
	Select r
		Case $00, $02, $04, $06, $08, $0A, $0C, $0E
			retval = VIC\MX[r >> 1]
		Case $01, $03, $05, $07, $09, $0B, $0D, $0F
			retval = VIC\MY[r >> 1]
		Case $10
			retval = VIC\MX_MSB
		Case $11
			; control register 1
			If VIC\Reg11Write
				retval = (VIC\CR1 & $7F) | ((VIC\CurrentRasterLine & $100) >> 1)
			EndIf
		Case $12
			retval = VIC\CurrentRasterLine & $FF
		Case $13
			; lightpen x-pos
		Case $14
			; lightpen y-pos
		Case $15
			retval = VIC\Sprite_Enabled
		Case $16
			; control register 2
			retval = VIC\CR2
		Case $17
			retval = VIC\Sprite_DoubleHeight			
		Case $18
			retval = VIC\Mem_VIC_Base
		Case $19
			retval = VIC\IRQ_Flag
		Case $1A
			retval = VIC\IRQ_Mask
		Case $1B
			retval = VIC\Sprite_Priority
		Case $1C
			retval = VIC\Sprite_Multicolor
		Case $1D
			retval = VIC\Sprite_DoubleWidth
		Case $1E
			retval = VIC\Sprite_Sprite_Collision
			VIC\Sprite_Sprite_Collision = 0
		Case $1F
			retval = VIC\Sprite_Backgr_Collision
			VIC\Sprite_Backgr_Collision = 0
		Case $20
			retval = VIC\Border_Color
		Case $21, $22, $23, $24
			retval = VIC\Backgr_Color[r - $21]
		Case $25, $26
			retval = VIC\Sprite_SharedColors[r - $25]
		Case $27, $28, $29, $2A, $2B, $2C, $2D, $2E
			retval = VIC\Sprite_Colors[r - $27]
		Default
 			retval = $FF
	EndSelect
	
	ProcedureReturn retval
EndProcedure

Procedure VIC_WriteRegister(r.a, v.a)
	Select r
		Case $00, $02, $04, $06, $08, $0A, $0C, $0E
			VIC\MX[r >> 1] = v
		Case $01, $03, $05, $07, $09, $0B, $0D, $0F
			VIC\MY[r >> 1] = v
		Case $10
			VIC\MX_MSB = v
		Case $11
			; control register 1
			VIC\CR1 = v
			
			VIC\VerticalScroll		= (VIC\CR1 & $07)
			VIC\ScreenRows25		= (VIC\CR1 & $08)
			VIC\Display_Enabled		= (VIC\CR1 & $10)
			VIC\GraphicMode			= ((VIC\CR1 & $60) | (VIC\CR2 & $10)) >> 4
			
			
			Protected new_Raster_IRQ.u
 			new_Raster_IRQ = (VIC\IRQ_RasterLine & $FF) | ((v & $80) << 1)
; 			If (VIC\IRQ_RasterLine <> new_Raster_IRQ) And (VIC\CurrentRasterLine = new_Raster_IRQ)
; 				VIC_RasterIRQ()
; 			EndIf
			VIC\IRQ_RasterLine			= new_Raster_IRQ
			
; 			If (VIC\CurrentRasterLine = $30) And (v & $10)
; 				VIC\Badline_Enabled = #True
; 			EndIf
; 			If (VIC\CurrentRasterLine >= $30) And (VIC\CurrentRasterLine <= $F7) And (VIC\VerticalScroll = (VIC\CurrentRasterLine & 7)) And VIC\Badline_Enabled
; 				VIC\Badline_status = #True
; 			Else
; 				VIC\Badline_status = #False
; 			EndIf
			
			VIC\Reg11Write = #True
		Case $12
			new_Raster_IRQ = (VIC\IRQ_RasterLine & $FF00) | v
; 			If (VIC\IRQ_RasterLine <> new_Raster_IRQ) And (VIC\CurrentRasterLine = new_Raster_IRQ)
; 				VIC_RasterIRQ()
; 			EndIf
			VIC\IRQ_RasterLine			= new_Raster_IRQ
		Case $15
			VIC\Sprite_Enabled = v
		Case $16
			; control register 2
			VIC\CR2 = v
			
			VIC\HorizontalScroll	= (VIC\CR2 & $07)
			VIC\ScreenCols40		= (VIC\CR2 & $08)
			VIC\GraphicMode			= ((VIC\CR1 & $60) | (VIC\CR2 & $10)) >> 4
		Case $17
			VIC\Sprite_DoubleHeight = v
		Case $18
			; memory pointers 
			VIC\Mem_Char		= (v & $0E) << 10;	bits ----321-
			VIC\Mem_Screen		= (v & $F0) <<  6;	bits 7654----
			VIC\Mem_Bitmap		= (v & $08) << 10;	bit  ----3---
			VIC\Mem_VIC_Base	= v | 1
		Case $19
			; interrupt request register: acknowledge interrupts by mask
			VIC\IRQ_Flag = (VIC\IRQ_Flag & (~v & $0F)); | $70			
			If VIC\IRQ_Flag & VIC\IRQ_Mask
				VIC\IRQ_Flag | $80
			Else
				CPU\IRQ_State & ~#VIC_IRQ
			EndIf
			
		Case $1A
			; interrupt control register
			VIC\IRQ_Mask = (v & $0F)
			If VIC\IRQ_Flag & VIC\IRQ_Mask
				VIC\IRQ_Flag | $80
				CPU\IRQ_State | #VIC_IRQ
			Else
				VIC\IRQ_Flag & $7F
				CPU\IRQ_State & ~#VIC_IRQ
			EndIf
		Case $1B
			VIC\Sprite_Priority = v
		Case $1C
			VIC\Sprite_Multicolor = v
		Case $1D
			VIC\Sprite_DoubleWidth = v
			; 		Case $1E
			; 			; sprite-sprite collision
			; 			VIC\Sprite_Sprite_Mask = v
			; 		Case $1F
			; 			; sprite-background collision
			; 			VIC\Sprite_Sprite_Mask = v
		Case $20
			VIC\Border_Color = v & $F
		Case $21, $22, $23, $24
			VIC\Backgr_Color[r - $21] = v & $F
		Case $25, $26
			VIC\Sprite_SharedColors[r - $25] = v & $F
		Case $27, $28, $29, $2A, $2B, $2C, $2D, $2E
			VIC\Sprite_Colors[r - $27] = v & $F
	EndSelect
EndProcedure

Procedure VIC_SetRasterCounter(v.u)
	VIC\CurrentRasterLine = v
	
	VIC\CR1 = (VIC\CR1 & $7F) | ((v >> 1) & $80)
	
; 	If (VIC\CurrentRasterLine = $30) And (VIC\Display_Enabled)
; 		VIC\Badline_Enabled = #True
; 	EndIf	
EndProcedure

Procedure.a VIC_GetScreenChar(column, row)
	Protected vicAddr.u = VIC\Mem_Screen + (row * #nColumns) + column
	ProcedureReturn MEMORY_VIC_ReadB(vicAddr)
EndProcedure

Procedure.a VIC_GetCharData(char, lin)
	Protected vicAddr.u = VIC\Mem_Char + (char << 3) + lin
	ProcedureReturn MEMORY_VIC_ReadB(vicAddr)
EndProcedure

Procedure.a VIC_GetBitmapData(column, row, lin)
	Protected vicAddr.u = VIC\Mem_Bitmap + ((row * #nColumns + column) << 3) + lin
	ProcedureReturn MEMORY_VIC_ReadB(vicAddr)
EndProcedure

Procedure.u VIC_GetSpritePointer(sprite)
	Protected vicAddr.u = VIC\Mem_Screen + #SpritePtrsOffset + sprite
	ProcedureReturn MEMORY_VIC_ReadB(vicAddr) * #SpriteSize
EndProcedure

Procedure.a VIC_GetCharColor(col.a, row.a)
	Protected ofs.u = ((row * #nColumns) + col)
	ProcedureReturn MEMORY\ColorRAM[ofs & $FFF] & $F
EndProcedure

Procedure VIC_SpriteX(n)
	Protected x = VIC\MX[n]
	
	If ISSET_BIT(VIC\MX_MSB, n)
		x + 256
	EndIf
	
	ProcedureReturn x
EndProcedure

Procedure.a VIC_UpdateSprCollision(sprite.a, x)
	Protected retval.a
	
	If VIC\SprCollision[x]
		VIC\Sprite_Sprite_Mask | (1 << sprite) | VIC\SprCollision[x]
		retval | %01
	EndIf		
	VIC\SprCollision[x] | (1 << sprite)
	
	If VIC\BgrCollision[x]
		VIC\Sprite_Backgr_Mask | (1 << sprite)
		retval | %10
	EndIf
	
	ProcedureReturn retval
EndProcedure

Procedure VIC_UpdateBgrCollision(x)
	VIC\BgrCollision[x] = 1
EndProcedure

Procedure VIC_DrawChar(x, y, charData.a, color.a)
	Protected i.a
	Protected xoffs.u = x + 7

	For i = 0 To 7
		If (xoffs >= 0) And (xoffs < #ScreenWidth)
			If (charData & 1)
				IO_DrawPixel(xoffs, y, color)
				VIC\Priority[xoffs] = #Priority_Front
				VIC_UpdateBgrCollision(xoffs)
			Else
				IO_DrawPixel(xoffs, y, VIC\Backgr_Color[0])
				VIC\Priority[xoffs] = #Priority_Back
			EndIf
		EndIf
		charData >> 1
		xoffs - 1
	Next
EndProcedure

Procedure VIC_DrawChar_MultiColor(x, y, charData.a, color.a)
	Protected i	
	Protected xoffs = x + 6
	Protected charColor.a, colorSource.a
	
	For i = 0 To 3
		If (xoffs >= 0) And (xoffs < #ScreenWidth)			
			colorSource = charData & $3
			
			If colorSource = %11
				charColor = color
			Else
				charColor = VIC\Backgr_Color[colorSource]
			EndIf
			
			IO_DrawPixel(xoffs,     y, charColor)
			IO_DrawPixel(xoffs + 1, y, charColor)
		
			If colorSource > 1
				VIC\Priority[xoffs    ] = #Priority_Front
				VIC\Priority[xoffs + 1] = #Priority_Front
				VIC_UpdateBgrCollision(xoffs)
				VIC_UpdateBgrCollision(xoffs + 1)
			Else
				VIC\Priority[xoffs    ] = #Priority_Back
				VIC\Priority[xoffs + 1] = #Priority_Back
			EndIf		

		EndIf
		xoffs - 2
		charData >> 2
	Next
EndProcedure

Procedure VIC_DrawChar_ExtBg(x, y, charData.a, color.a, Backgr_Color.a)
	Protected xoffs = x + 7

	Protected i
	For i = 0 To 7
		If (xoffs >= 0) And (xoffs < #ScreenWidth)
			; draw pixel
			If (charData & 1)
				IO_DrawPixel(xoffs, y, color)
				VIC\Priority[xoffs] = #Priority_Front
				VIC_UpdateBgrCollision(xoffs)
			Else
				IO_DrawPixel(xoffs, y, Backgr_Color)
				VIC\Priority[xoffs] = #Priority_Back
			EndIf
		EndIf
		charData >> 1
		xoffs - 1
	Next
EndProcedure

Procedure VIC_DrawRaster_CharMode(columnNr, y, xScroll.a, yScroll.a)
	Protected x			= (columnNr << 3) + 92 + xScroll
  	Protected lineNr	= y - 47 - yScroll
	Protected row		= (lineNr >> 3)
	Protected char_row	= (lineNr % 8)
   	Protected c.a		= VIC_GetScreenChar(columnNr, row)
	Protected color.a	= VIC_GetCharColor(columnNr, row)
	Protected d.a
	
	IO_DrawLine(x, y, 8, VIC\Backgr_Color)
	
	If VIC\GraphicMode = #EXTENDED_BACKGROUND_COLOR
		d 	= VIC_GetCharData(c & $3F, char_row)
		VIC_DrawChar_ExtBg(x, y, d, color & $7, VIC\Backgr_Color[c >> 6] & $7)
	ElseIf (VIC\GraphicMode = #MULTICOLOR_TEXT)
		d	= VIC_GetCharData(c, char_row)
		If (color & $8)
			VIC_DrawChar_MultiColor(x, y, d, color & $7)
		Else
 			VIC_DrawChar(x, y, d, color & $7)
		EndIf
	Else
		d	= VIC_GetCharData(c, char_row)
  		VIC_DrawChar(x, y, d, color & $F)
	EndIf
EndProcedure

Procedure VIC_DrawBitmap(x, y, d.a, color.a)
	Protected fgc.a = (color >> 4) & $0F
	Protected bgc.a = color & $0F
	Protected xoffs = x
	
	Protected i
	For i = 0 To 7
		If (xoffs >= 0) And (xoffs < #ScreenWidth)
			; draw pixel
			If (d & 1)
				IO_DrawPixel(xoffs, y, fgc)
				VIC\Priority[xoffs] = #Priority_Front
				VIC_UpdateBgrCollision(xoffs)
			Else
				IO_DrawPixel(xoffs, y, bgc)
				VIC\Priority[xoffs] = #Priority_Back
			EndIf
		EndIf
		d >> 1
		xoffs - 1
	Next
EndProcedure

Procedure VIC_DrawBitmap_MultiColor(x, y, d.a, scolor.a, rcolor.a)
	Protected i
	For i = 0 To 3
		Protected xoffs = x + 6 - (i * 2)
		If (xoffs >= 0) And (xoffs < #ScreenWidth)
			Protected color.a
			Protected colorSource.a = (d >> (i << 1)) & $3
			Select colorSource
				Case %00: color = VIC\Backgr_Color[0]
				Case %01: color = (scolor >> 4) & $F
				Case %10: color = scolor & $F
				Case %11: color = rcolor & $F
			EndSelect
			
			IO_DrawPixel(xoffs,     y, color)
			IO_DrawPixel(xoffs + 1, y, color)
			
			If colorSource > 1
				VIC\Priority[xoffs    ] = #Priority_Front
				VIC\Priority[xoffs + 1] = #Priority_Front
				VIC_UpdateBgrCollision(xoffs)
				VIC_UpdateBgrCollision(xoffs + 1)
			Else
				VIC\Priority[xoffs    ] = #Priority_Back
				VIC\Priority[xoffs + 1] = #Priority_Back
			EndIf
		EndIf
	Next
EndProcedure

Procedure VIC_DrawRaster_BitmapMode(columnNr, y)
	Protected x				= (columnNr << 3) + 92
	Protected lineNr 		= y - 50
	Protected row			= lineNr >> 3
	Protected bitmap_row	= lineNr % 8
	Protected d.a, scolor.a, rcolor.a
	
	If y < #FirstVisibleLine Or y > #LastVisibleLine
		ProcedureReturn
	EndIf	
	
	d		= VIC_GetBitmapData(columnNr, row, bitmap_row)
	scolor	= VIC_GetScreenChar(columnNr, row)
	If VIC\GraphicMode = #STANDARD_BITMAP
  		VIC_DrawBitmap(x + 8, y, d, scolor)
	ElseIf VIC\GraphicMode = #MULTICOLOR_BITMAP
		rcolor	= VIC_GetCharColor(columnNr, row)
		VIC_DrawBitmap_MultiColor(x, y, d, scolor, rcolor)
	EndIf
EndProcedure

Procedure VIC_DrawSprite(x, y, sprite, row)
	Protected spriteData.a
	Protected xoffs
	Protected w, i, j
	Protected spriteOffs.u		= VIC_GetSpritePointer(sprite)
	Protected spritePriority	= VIC_SpritePriority(sprite)
	Protected color				= VIC\Sprite_Colors[sprite]
	
	If VIC_IsDoubleWidthSprite(sprite)
		For i = 0 To 2
			spriteData = MEMORY_VIC_ReadB(spriteOffs + row * 3 + i)
			xoffs = x + ((i << 3) + 7) << 1	
			For j = 0 To 7
				If spriteData & 1
					If (spritePriority = #Priority_Front) Or (VIC\Priority[xoffs] = #Priority_Back)
						If (xoffs >= 0) And (xoffs < #ScreenWidth)		
							IO_DrawPixel(xoffs    , y, color)
							IO_DrawPixel(xoffs + 1, y, color)
 							VIC\Priority[xoffs    ] = #Priority_Front
 							VIC\Priority[xoffs + 1] = #Priority_Front
;  							VIC_UpdateBgrCollision(xoffs    )
;  							VIC_UpdateBgrCollision(xoffs + 1)
 						EndIf
 						VIC_UpdateSprCollision(sprite, xoffs + 0)
 						VIC_UpdateSprCollision(sprite, xoffs + 1)
					EndIf
				EndIf
				spriteData >> 1
				xoffs - 2
			Next
		Next
		
	Else
		
		For i = 0 To 2
			spriteData = MEMORY_VIC_ReadB(spriteOffs + row * 3 + i)
			xoffs = x + (i << 3) + 7
			For j = 0 To 7
				If spriteData & 1
					If (spritePriority = #Priority_Front) Or (VIC\Priority[xoffs] = #Priority_Back)
						If (xoffs >= 0) And (xoffs < #ScreenWidth)
							color = VIC\Sprite_Colors[sprite]
 							IO_DrawPixel(xoffs, y, color)
 							VIC\Priority[xoffs] = #Priority_Front
;  							VIC_UpdateBgrCollision(xoffs)
 						EndIf
 					EndIf
 					VIC_UpdateSprCollision(sprite, xoffs)
				EndIf
				spriteData >> 1
				xoffs - 1
			Next
		Next
		
	EndIf
EndProcedure

Procedure VIC_DrawSprite_Multicolor(x, y, sprite, row)
	Protected i, j
	Protected color.a, colorSource.a
	Protected xoffs
	Protected ofs.u				= VIC_GetSpritePointer(sprite)
	Protected spritePriority	= VIC_SpritePriority(sprite)
	
	If VIC_IsDoubleWidthSprite(sprite)		
		For i = 0 To 2
			colorSource = MEMORY_VIC_ReadB(ofs + (row * 3) + i)
			xoffs = x + ((i << 3) + 8) << 1			
			For j = 0 To 3
				If colorSource
					Select colorSource & $3
						Case 0: color = 0
						Case 1:	color = VIC\Sprite_SharedColors[0]
						Case 2:	color = VIC\Sprite_Colors[sprite]
						Case 3:	color = VIC\Sprite_SharedColors[1]
					EndSelect
					If colorSource & $3
						If (spritePriority = #Priority_Front) Or (VIC\Priority[xoffs] = #Priority_Back)
							If (xoffs >= 0) And (xoffs < #ScreenWidth)
								IO_DrawPixel(xoffs + 0, y, color)
								IO_DrawPixel(xoffs + 1, y, color)
								IO_DrawPixel(xoffs + 2, y, color)
								IO_DrawPixel(xoffs + 3, y, color)
								VIC\Priority[xoffs    ] = #Priority_Front
								VIC\Priority[xoffs + 1] = #Priority_Front
								VIC\Priority[xoffs + 2] = #Priority_Front
								VIC\Priority[xoffs + 3] = #Priority_Front
							EndIf
						EndIf
						If colorSource & 2
							VIC_UpdateSprCollision(sprite, xoffs + 0)
							VIC_UpdateSprCollision(sprite, xoffs + 1)
							VIC_UpdateSprCollision(sprite, xoffs + 2)
							VIC_UpdateSprCollision(sprite, xoffs + 3)
						EndIf

					EndIf
				EndIf
				colorSource >> 2
				xoffs - 4
			Next
		Next
		
	Else
		
		For i = 0 To 2			
			colorSource = MEMORY_VIC_ReadB(ofs + (row * 3) + i)
			xoffs = x + (i << 3) + 7
			For j = 0 To 3
				If colorSource
					Select colorSource & $3
						Case 0: color = 0
						Case 1:	color = VIC\Sprite_SharedColors[0]
						Case 2:	color = VIC\Sprite_Colors[sprite]
						Case 3:	color = VIC\Sprite_SharedColors[1]
					EndSelect
					If colorSource & $3
						If (spritePriority = #Priority_Front) Or (VIC\Priority[xoffs] = #Priority_Back)
							If (xoffs >= 0) And (xoffs < #ScreenWidth)
								IO_DrawPixel(xoffs    , y, color)
								IO_DrawPixel(xoffs + 1, y, color)
								VIC\Priority[xoffs + 0] = #Priority_Front
								VIC\Priority[xoffs + 1] = #Priority_Front
							EndIf
						EndIf
						If colorSource & 2
							VIC_UpdateSprCollision(sprite, xoffs + 0)
							VIC_UpdateSprCollision(sprite, xoffs + 1)
						EndIf
					EndIf
				EndIf
				colorSource >> 2
				xoffs - 2
			Next
		Next 
	EndIf
EndProcedure

Procedure VIC_DrawRaster_Sprites(lineNr)
	If VIC\DrawLine = 0
		ProcedureReturn
	EndIf
		
	VIC\Sprite_Backgr_Mask = 0
	VIC\Sprite_Sprite_Mask = 0
	
	If VIC\Sprite_Enabled = 0
		ProcedureReturn
	EndIf
	
	Protected y = lineNr; - 3
	Protected sp_y = y
	Protected sprite.a, height.u
	Protected row
	
	For sprite = 7 To 0 Step - 1
		If VIC_IsSpriteEnabled(sprite)
			If VIC_IsDoubleHeightSprite(sprite)
				height = #SpriteHeight << 1
			Else 
				height = #SpriteHeight
			EndIf
			
			;check If the sprite is visible
			If (sp_y >= VIC\MY[sprite]) And
			   (sp_y < (VIC\MY[sprite] + height))
				
				If VIC_IsDoubleHeightSprite(sprite)
					row = (sp_y - VIC\MY[sprite]) >> 1
				Else
					row = (sp_y - VIC\MY[sprite])
				EndIf
				
				If VIC_IsMulticolorSprite(sprite)
					VIC_DrawSprite_Multicolor(69 + VIC_SpriteX(sprite), y, sprite, row)
				Else
					VIC_DrawSprite(69 + VIC_SpriteX(sprite), y, sprite, row)
				EndIf
			EndIf
		EndIf
	Next
	
	If VIC\Sprite_Sprite_Collision = 0
		VIC\Sprite_Sprite_Collision | VIC\Sprite_Sprite_Mask
		If VIC\Sprite_Sprite_Collision
			VIC\IRQ_Flag | $4
			If VIC\IRQ_Mask & $4
				VIC\IRQ_Flag | $80					
				CPU\IRQ_State | #VIC_IRQ
			EndIf
		EndIf
	Else
		VIC\Sprite_Sprite_Collision | VIC\Sprite_Sprite_Mask
	EndIf
	
	If VIC\Sprite_Backgr_Collision = 0
		VIC\Sprite_Backgr_Collision | VIC\Sprite_Backgr_Mask
		If VIC\Sprite_Backgr_Collision
			VIC\IRQ_Flag | $2
			If VIC\IRQ_Mask & $2
				VIC\IRQ_Flag | $80
				CPU\IRQ_State | #VIC_IRQ
			EndIf
		EndIf
	Else
		VIC\Sprite_Backgr_Collision | VIC\Sprite_Backgr_Mask
	EndIf
EndProcedure

Procedure VIC_DrawBorder(y)
	If VIC\DrawLine = 0
		ProcedureReturn
	EndIf

	Protected borderLeft, borderLeftSize
	Protected borderRight, borderRightSize
	Protected borderTop, BorderBottom
	
	If y < #FirstVisibleLine Or y > #LastVisibleLine
		ProcedureReturn
 	EndIf
	
	If VIC\ScreenRows25
		borderTop		= 50
		BorderBottom	= 249
	Else
		borderTop		= 54
		BorderBottom	= 245
	EndIf
	
	If VIC\ScreenCols40
		borderLeft		= 52
		borderLeftSize	= 40
		borderRight		= 412
		borderRightSize	= 43
	Else
		borderLeft		= 52
		borderLeftSize	= 48
		borderRight		= 404
		borderRightSize	= 51
	EndIf	
	
	If (y < borderTop) Or (y > BorderBottom)
		IO_DrawLine(52,				y, 403, 			VIC\Border_Color)
	Else
		IO_DrawLine(borderLeft,		y, borderLeftSize,	VIC\Border_Color)
 		IO_DrawLine(borderRight,	y, borderRightSize,	VIC\Border_Color)
	EndIf
EndProcedure

Procedure VIC_DrawGraphics(columnNr, lineNr.u, drawLine.a, xScroll.a, yScroll.a)
	If drawLine = #False Or (VIC\Cycle < 16) Or (VIC\Cycle > 55); Or (lineNr < 50) Or (lineNr > 249)
		ProcedureReturn
	EndIf

	Select VIC\GraphicMode
		Case #STANDARD_TEXT, #MULTICOLOR_TEXT, #EXTENDED_BACKGROUND_COLOR
			VIC_DrawRaster_CharMode(columnNr - 16, lineNr, xScroll, yScroll)
		Case #STANDARD_BITMAP, #MULTICOLOR_BITMAP
			VIC_DrawRaster_BitmapMode(columnNr - 16, lineNr)
; 		Default
;  			C64\ErrorMessage	= "VIC: Unsupported Graphicmode: " + Str(VIC\GraphicMode) + "."
;  			C64\State			| #C64_State_Error
	EndSelect
EndProcedure

Procedure VIC_Emulate()
	Static xScroll, yScroll
	
	If VIC\CurrentRasterLine = $30
		VIC\Badline_Enabled = VIC\Display_Enabled
	EndIf
    If (VIC\CurrentRasterLine >= $30) And (VIC\CurrentRasterLine <= $F7) And (VIC\VerticalScroll = (VIC\CurrentRasterLine & 7)) And VIC\Badline_Enabled
        VIC\Badline_status	= #True
        VIC\Display_status	= #True
    Else
    	VIC\Badline_status	= #False
    EndIf
	
	Select VIC\Cycle
		Case 1
			If (VIC\CurrentRasterLine = #ScreenHeight - 1)
				IO_RefreshScreen()
				VIC\CurrentRasterLine = 0
				VIC_SetRasterCounter(0)
   			Else
				VIC_SetRasterCounter(VIC\CurrentRasterLine + 1)
				If (VIC\CurrentRasterLine >= #FirstVisibleLine) And (VIC\CurrentRasterLine <= #LastVisibleLine)
					VIC\DrawLine = #True
				Else
					VIC\DrawLine = #False
				EndIf
			EndIf
			
			; clear buffers
			FillMemory(@VIC\Priority, SizeOf(VIC\Priority), 0, #PB_Byte)
			FillMemory(@VIC\SprCollision, SizeOf(VIC\SprCollision), 0, #PB_Byte)
			FillMemory(@VIC\BgrCollision, SizeOf(VIC\BgrCollision), 0, #PB_Byte)			
		Case 2
			If VIC\CurrentRasterLine = VIC\IRQ_RasterLine
 				VIC_RasterIRQ()
 			EndIf
  		Case 11 To 61
  			If VIC\Cycle = 14
  				xScroll = VIC\HorizontalScroll
  				yScroll = VIC\VerticalScroll
  			EndIf
  			
			VIC_DrawGraphics(VIC\Cycle, VIC\CurrentRasterLine, VIC\DrawLine, xScroll, yScroll)
		Case 63
			VIC_DrawRaster_Sprites(VIC\CurrentRasterLine)
  			VIC_DrawBorder(VIC\CurrentRasterLine)
		Case 65
			VIC\Cycle = 0
	EndSelect
	
	VIC\Cycle + 1

	ProcedureReturn #True
EndProcedure; 
; IDE Options = PureBasic 5.72 (Windows - x64)
; CursorPosition = 770
; FirstLine = 716
; Folding = ------
; EnableXP